import Image
import os

class Saver():
    def __init__(self, extension, output_path, filename_prefix):
        self.output_path = output_path
        self.filename_prefix = filename_prefix
        self.last_index = 0
        self.extension = extension

        if self.filename_prefix and not self.filename_prefix.endswith("-"):
            self.filename_prefix += "-"


    def next_filename(self):
        self.last_index += 1
        return "%s%d.%s" % (self.filename_prefix, self.last_index, self.extension)

    def save(self, sprite_sheet):
        result = Image.new('RGBA', sprite_sheet.size, (0xff, 0xff, 0xff, 1))

        for image, position in sprite_sheet:
            result.paste(image, position)

        sprite_sheet.filename = self.next_filename()

        filename = os.path.join(self.output_path, sprite_sheet.filename)
        self.save_file(result, filename)

        sprite_sheet.filesize = os.path.getsize(filename)

    def save_file(self, image, filename):
        raise "This method should be overridden in inherited classes"



import ImageChops
import os
import Image
import re
from math import sqrt

class SpriteSheet():
    def __init__(self):
        self.sprites = []
        self.used_area = 0
        self.size = (0, 0)
        self.filename = ''
        self.filesize = 0

    def get_metadata(self):
        metadata = {}
        for sprite in self.sprites:
            metadata[sprite['path']] = {
                'position': sprite['position'],
                'sprite_sheet': self.filename,
            }
        return metadata

    def add_sprite(self, sprite, position):
        sprite['position'] = position
        self.sprites.append(sprite)
        self.used_area += sprite['area']

    def __iter__(self):
        for x in self.sprites:
            yield (x['image'], x['position'])

    def __len__(self):
        return len(self.sprites)

class ArtPacker():
    def __init__(self, input_path, metadata_saver, image_saver, resource_packer, padding=0, duplicates_threshold=None, input_regex=None, verbose=False):
        self.input_path = input_path
        self.metadata_saver = metadata_saver
        self.image_saver = image_saver
        self.resource_packer = resource_packer
        self.verbose = verbose
        self.padding = padding
        self.input_regex = input_regex
        self.duplicates_threshold = duplicates_threshold

    def generate(self):
        if self.verbose:
            print "Reading image resources"

        images, duplicates = self.read_sprites()

        if not images:
            if self.verbose:
                print "No images was found"
            raise SystemExit


        if self.verbose:
            print "Packing %d images to sprite-sheets" % len(images)

        source_filesize = reduce(lambda total, image: image['filesize'] + total, images, 0)

        last_sprite_sheet = 1
        #Some statistics
        total_used_area = 0
        total_sheet_area = 0
        total_filesize = 0

        metadata = {}
        sprite_sheets = []

        while images:
            sprite_sheet, images = self.resource_packer.pack(images)
            self.image_saver.save(sprite_sheet)

            total_used_area += sprite_sheet.used_area
            sheet_area = sprite_sheet.size[0] * sprite_sheet.size[1]
            total_sheet_area += sheet_area
            total_filesize += sprite_sheet.filesize
            last_sprite_sheet += 1
            sprite_sheets.append(sprite_sheet.filename)
            metadata.update(sprite_sheet.get_metadata())
            if self.verbose:
                overhead = 100.0 - (100.0 * sprite_sheet.used_area / sheet_area)
                print "%s %d images was packed with pixel overhead of %.2f%%" % (sprite_sheet.filename, len(sprite_sheet), overhead)

        for image in duplicates:
            metadata[image['path']] = metadata[image['duplicate_of']]

        self.metadata_saver.save({'sheets': sprite_sheets, 'sprites': metadata})
        if self.verbose:
            total_overhead = 100.0 - (100.0 * total_used_area / total_sheet_area)
            print "Done!"
            print "Total %d images was packed to %d sprite-sheets with pixel overhead of %.2f%%" % (len(metadata), len(sprite_sheets), total_overhead)
            print "Total size of unpacked resources is %.2f Mb" % (source_filesize / 1024.0 / 1024.0)
            print "Total size of sprite sheets is %.2f Mb" % (total_filesize / 1024.0 / 1024.0)

    def read_sprites(self):
        images = []

        for dirname, dirnames, filenames in os.walk(self.input_path):
            for filename in filenames:
                if self.input_regex and not re.match(self.input_regex, filename):
                    continue

                file_path = os.path.join(dirname, filename)

                try:
                    image = Image.open(file_path)
                except Exception, e:
                    continue

                #Crop by alpha channel if any
                if 'A' in image.getbands():
                    bbox = image.split()[image.getbands().index('A')].getbbox()
                else:
                    bbox = image.getbbox()

                #check Image is not empty
                if not bbox:
                    continue

                #Crop image to non-transparent area
                image = image.crop(bbox)

                #Add padding to the image
                image = self.add_padding(image)

                images.append({
                    'image': image,
                    'path': os.path.relpath(file_path, self.input_path),
                    'size': image.size,
                    'area': image.size[0] * image.size[1],
                    'width': image.size[0],
                    'height': image.size[1],
                    'filesize': os.path.getsize(file_path)
                })

        return self.filter_duplicates(images)

    def add_padding(self, image):
        if self.padding <= 0:
            return image

        #We use this hand-written code because ImageOps.expand doesn't add transparent background
        width = self.padding + image.size[0] + self.padding
        height = self.padding + image.size[1] + self.padding
        result = Image.new("RGBA", (width, height), (0xff, 0xff, 0xff, 1))
        result.paste(image, (self.padding, self.padding))
        return result

    def filter_duplicates(self, images):
        if self.duplicates_threshold is None:
            return images, []

        result = []
        duplicates = []
        for image in images:
            is_duplicate = False
            for candidate in result:
                if image_match(candidate['image'], image['image'], self.duplicates_threshold):
                    is_duplicate = True
                    break
            if not is_duplicate:
                result.append(image)
            else:
                del image['image']
                image['duplicate_of'] = candidate['path']
                duplicates.append(image)

        return result, duplicates

def image_match(image_a, image_b, threshold):
    if image_a.size != image_b.size or image_a.getbands() != image_b.getbands():
        return False

    rms = []
    for band_a, band_b in zip(image_a.split(), image_b.split()):
        hist = ImageChops.difference(band_a, band_b).histogram()
        squares = (value * (i ** 2.0) for i, value in enumerate(hist))
        area = image_a.size[0] * image_a.size[1]
        rms.append(sqrt(sum(squares) / area))

    rms = sqrt(reduce(lambda score, rms: rms ** 2 + score, rms, 0))

    return rms <= threshold




import os
from artpacker.artpacker import SpriteSheet

class PackNode(object):
    """
    Creates an area which can recursively pack other areas of smaller sizes into itself.
    """
    def __init__(self, area):
        #if tuple contains two elements, assume they are width and height, and origin is (0,0)
        if len(area) == 2:
            area = (0, 0, area[0], area[1])
        self.area = area

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, str(self.area))

    def get_width(self):
        return self.area[2] - self.area[0]

    width = property(fget=get_width)

    def get_height(self):
        return self.area[3] - self.area[1]

    height = property(fget=get_height)

    def insert(self, area):
        if hasattr(self, 'child'):
            a = self.child[0].insert(area)

            if a is None:
                return self.child[1].insert(area)

            return a

        area = PackNode(area)
        if area.width <= self.width and area.height <= self.height:
            self.child = [None, None]
            self.child[0] = PackNode((self.area[0] + area.width, self.area[1], self.area[2], self.area[1] + area.height))
            self.child[1] = PackNode((self.area[0], self.area[1] + area.height, self.area[2], self.area[3]))
            return PackNode((self.area[0], self.area[1], self.area[0] + area.width, self.area[1] + area.height))

class SimplePacker():
    def __init__(self, max_width, max_height):
        self.max_width =  max_width
        self.max_height = max_height

    def pack(self, images):
        result = SpriteSheet()
        source_images = sorted(images, key=lambda item: item['area'], reverse=True)
        output_size = self.adjust_max_size(source_images)

        tree = PackNode(output_size)

        missed_images = []
        for image in source_images:
            position = tree.insert(image['size'])
            if position is None:
                missed_images.append(image)
            else:
                result.size = (max(result.size[0], position.area[2]), max(result.size[1], position.area[3]))
                result.add_sprite(image, position.area)

        return result, missed_images

    def adjust_max_size(self, images):
        #The largest image must fit the sprite sheet
        max_width = max(self.max_width, max(map(lambda a: a['width'], images)))
        max_height = max(self.max_height, max(map(lambda a: a['height'], images)))
        return max_width, max_height




import os
from artpacker import ArtPacker
from optparse import OptionParser

from metadata.json import JSONMetadataSaver
from metadata.dummy import DummyMetadataSaver
from packer.simple import SimplePacker
from saver.png import PNGSaver
from saver.jpeg import JPEGSaver

AVAILABLE_FORMATS = ('png', 'jpg')

def main():
    parser = OptionParser(usage="%prog [options]")

    parser.add_option("-i", "--input-path",
        dest="input",
        help="Path to image folder, required")

    parser.add_option("-o", "--output-path",
        dest="output",
        default=".",
        help="Path to output folder. Default '.'")

    parser.add_option("--metadata-format",
        dest="metadata_format",
        default="json",
        help="Metadata file format. Should be one of 'json', 'dummy'. Default 'json'")

    parser.add_option("--metadata-filename",
        dest="metadata_filename",
        help="Metadata file name")

    parser.add_option("--sprite-sheet-prefix",
        dest="prefix",
        default='',
        help="Prefix of sprite sheet file names.")

    parser.add_option("--width",
        dest="width", default=1024, type="int",
        help="Sprite sheet image width. Default 1024")

    parser.add_option("--height",
        dest="height", default=1024, type="int",
        help="Sprite sheet image height. Default 1024")

    parser.add_option("--padding",
        dest="padding", default=0, type="int",
        help="Padding between images in the sprite sheet")

    parser.add_option("--duplicates-threshold",
        dest="duplicates_threshold", default=None, type="float",
        help="Threshold for automatic duplicates filtering. Reasonable values are from 0 to 25. 0 is exact match")

    parser.add_option("--input-regex",
        dest="input_regex", default=None,
        help="Regular expression for filtering input files")

    names = ", ".join(map(lambda a: "'%s'" % a, AVAILABLE_FORMATS))
    parser.add_option("--output-format",
        dest="output_format", default='png',
        help="Output files format. Should be one of %s" % names)

    parser.add_option("--jpeg-quality",
        dest="jpeg_quality", default='85', type="int",
        help="Quality of JPEG file. Reasonable values 70-100")

    parser.add_option("--jpeg-progressive",
        action='store_true',
        dest="jpeg_progressive", default='False',
        help="Produce progressive JPEG file")

    parser.add_option("-q", "--quiet",
        action='store_false',
        dest="verbose", default=True,
        help="Don't print status messages to stdout")

    (options, args) = parser.parse_args()

    if not options.input:
        print "--input-path parameter is required"
        raise SystemExit

    if options.output_format not in AVAILABLE_FORMATS:
        print "--output-format %s is not supported" % options.output_format
        raise SystemExit


    packer = SimplePacker(
        max_width=options.width,
        max_height=options.height)

    ArtPacker(
        input_path=options.input,
        metadata_saver=get_metadata_saver(options),
        image_saver=get_image_saver(options),
        input_regex=options.input_regex,
        resource_packer=packer,
        padding=options.padding,
        duplicates_threshold=options.duplicates_threshold,
        verbose=options.verbose).generate()

def get_image_saver(options):
    if options.output_format == 'jpg':
        return JPEGSaver(
            output_path=options.output,
            filename_prefix=options.prefix,
            progressive=options.jpeg_progressive,
            quality=options.jpeg_quality)
    elif options.output_format == 'png':
        return PNGSaver(
            output_path=options.output,
            filename_prefix=options.prefix)

def get_metadata_saver(options):
    if options.metadata_format == 'json':
        if options.metadata_filename:
            metadata_filename = options.metadata_filename
        elif options.prefix:
            metadata_filename = options.prefix + '.json'
        else:
            metadata_filename = 'metadata.json'
        return JSONMetadataSaver(os.path.join(options.output, metadata_filename))
    else:
        return DummyMetadataSaver()



import ImageFile
from artpacker.saver import Saver

class JPEGSaver(Saver):
    def __init__(self, output_path, filename_prefix, progressive=False, quality=85):
        Saver.__init__(self, 'jpg', output_path, filename_prefix)
        self.progressive = progressive
        self.quality = quality
        self.output_path = output_path

    def save_file(self, image, filename):
        #Increase buffer to allow large images processing
        ImageFile.MAXBLOCK = image.size[0] * image.size[1]
        image.save(filename, 'JPEG', quality=self.quality, optimize=True, progressive=self.progressive)




try: 
    import simplejson as json
except ImportError: 
    import json

BANNER = "/* This file was automatically generated by art-packer and shouldn't be changed manually */\n"

class JSONMetadataSaver:
    def __init__(self, filename):
         self.filename = filename
  
    def save(self, metadata):
        content = BANNER + json.dumps(metadata, indent='  ')
        with open(self.filename, 'w') as fh:
            fh.write(content)




from artpacker.saver import Saver

class PNGSaver(Saver):
    def __init__(self, output_path, filename_prefix):
        Saver.__init__(self, 'png', output_path, filename_prefix)

    def save_file(self, image, filename):
        image.save(filename, optimize=True)